/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief SQLite Database Operations
 *
 * \author Steve Sokol <ssokol@sokol-associates.com>
 */
 
#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 46778 $")

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <libgen.h>
#include <sqlite3.h>

#include "asterisk/lock.h"
#include "asterisk/channel.h"
#include "asterisk/logger.h"
#include "asterisk/options.h"
#include "asterisk/file.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/manager.h"
#include "asterisk/cli.h"
#include "asterisk/utils.h"
#include "asterisk/config.h"
#include "asterisk/app.h"
#include "asterisk/res_sqlite3.h"

static char *config = "func_sqlite.conf";

enum {
	OPT_ESCAPECOMMAS =	(1 << 0),
} sqlite_option_flags;

struct acf_sqlite_query {
	AST_LIST_ENTRY(acf_sqlite_query) list;
	char file[PATH_MAX];
	char sql_read[2048];
	char sql_write[2048];
	unsigned int flags;
	struct ast_custom_function *acf;
};

AST_LIST_HEAD_STATIC(sqlite_queries, acf_sqlite_query);

/* 
 *	This is implemented as a res because the locking mechanism in SQLite3 is not 
 *	necessarily safe.  The res also prevents us from having to constantly open 
 *	and close files, as it maintains a pool of connections.
 */

static int sqlite_busy(void *cbp, int tries) {
	sleep(20);
	if (tries > 5)
		return 0;
		
	return -1;
}

static int sqlite_manager_callback(void *ms, int argc, char **argv, char **azColName) {
	int i, rcount;
	struct sqlite_manager_cb *smc = ms;
	struct mansession *s = smc->session;
	rcount = smc->record_count;
	 
	if (rcount == 0) {
		astman_append(s, "Response: Success\r\n");
	}
	
	astman_append(s, "Category-%06d: %d\r\n", rcount, rcount);	
	for(i=0; i<argc; i++){
		astman_append(s, "Line-%06d-%06d: %s=%s\r\n", rcount, i, azColName[i], argv[i] ? argv[i] : "");
	}
	smc->record_count++;
	return 0;
}


static int sqlite_function_callback(void *cbp, int argc, char **argv, char **azColName) {
	int i = 0, buflen = 0, x, rcount;
	struct sqlite_function_cb *sfc = cbp;
	char *buf = sfc->buf;
	size_t len = sfc->len;
	rcount = sfc->record_count;
	char *coldata = NULL;
	
	if (rcount == 0) {
		sfc->column_count = argc;
	}

	/* Set the buffer to null */
	*buf = '\0';

	for (x = 0; x < argc; x++) {
		buflen = strlen(buf);
		coldata = ast_strdup(argv[x]);

		/* Copy data, encoding '\' and ',' for the argument parser */
		for (i = 0; i < strlen(coldata); i++) {
			if (sfc->escapecommas && (coldata[i] == '\\' || coldata[i] == ',')) {
				buf[buflen++] = '\\';
			}
			buf[buflen++] = coldata[i];

			if (buflen >= len - 2)
				break;

			if (coldata[i] == '\0')
				break;
		}

		buf[buflen] = ',';
		buf[buflen + 1] = '\0';
	}
	/* Trim trailing comma */
	buf[buflen] = '\0';
	
	sfc->record_count++;
	return 0;
}


static int acf_sqlite_write(struct ast_channel *chan, char *cmd, char *s, const char *value)
{
	struct sqlite3 *db;
	struct acf_sqlite_query *query;
	char *t, buf[2048]="", varname[15];
	int i, rc, ret = 0;
	char *file;
	char dbpath[PATH_MAX];
	char *sql_error = NULL;	
	
	pbx_builtin_setvar_helper(chan, "SQL_ERROR", NULL);
	
	AST_DECLARE_APP_ARGS(values,
		AST_APP_ARG(field)[100];
	);
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(field)[100];
	);
	
	AST_LIST_LOCK(&sqlite_queries);
	AST_LIST_TRAVERSE(&sqlite_queries, query, list) {
		if (!strcmp(query->acf->name, cmd)) {
			break;
		}
	}

	if (!query) {
		ast_log(LOG_ERROR, "No such function '%s'\n", cmd);
		pbx_builtin_setvar_helper(chan, "SQL_ERROR", "No matching function.");
		AST_LIST_UNLOCK(&sqlite_queries);
		return -1;
	}

	file = ast_strdup(query->file);

	if (file[0] == '/') {
		/* full path */
		strncpy(dbpath, file, sizeof(file));
	} else {
		sprintf(dbpath, "/etc/asterisk/%s", file);
	}

	if (sqlite3_open(dbpath, &db)) {
		ast_log(LOG_ERROR, "Failed to open database '%s' (check func_sqlite.conf)\n", dbpath);
		pbx_builtin_setvar_helper(chan, "SQL_ERROR", "Failed to open database.");
		AST_LIST_UNLOCK(&sqlite_queries);
		return -1;
	}

	sqlite3_busy_handler(db, sqlite_busy, NULL);
	
	/* Parse our arguments */
	t = value ? ast_strdupa(value) : "";

	if (!s || !t) {
		ast_log(LOG_ERROR, "Out of memory\n");
		AST_LIST_UNLOCK(&sqlite_queries);
		return -1;
	}

	AST_STANDARD_APP_ARGS(args, s);
	for (i = 0; i < args.argc; i++) {
		snprintf(varname, sizeof(varname), "ARG%d", i + 1);
		pbx_builtin_pushvar_helper(chan, varname, args.field[i]);
	}

	/* Parse values, just like arguments */
	/* Can't use the pipe, because app Set removes them */
	AST_NONSTANDARD_APP_ARGS(values, t, ',');
	for (i = 0; i < values.argc; i++) {
		snprintf(varname, sizeof(varname), "VAL%d", i + 1);
		pbx_builtin_pushvar_helper(chan, varname, values.field[i]);
	}

	/* Additionally set the value as a whole (but push an empty string if value is NULL) */
	pbx_builtin_pushvar_helper(chan, "VALUE", value ? value : "");

	pbx_substitute_variables_helper(chan, query->sql_write, buf, sizeof(buf) - 1);

	/* Restore prior values */
	for (i = 0; i < args.argc; i++) {
		snprintf(varname, sizeof(varname), "ARG%d", i + 1);
		pbx_builtin_setvar_helper(chan, varname, NULL);
	}

	for (i = 0; i < values.argc; i++) {
		snprintf(varname, sizeof(varname), "VAL%d", i + 1);
		pbx_builtin_setvar_helper(chan, varname, NULL);
	}
	pbx_builtin_setvar_helper(chan, "VALUE", NULL);

	AST_LIST_UNLOCK(&sqlite_queries);

	rc = sqlite3_exec(db, buf, NULL, NULL, &sql_error);
	if (rc != SQLITE_OK) {
		ast_log(LOG_ERROR, "Query failed to execute: %s\n", sql_error);
		pbx_builtin_setvar_helper(chan, "SQL_ERROR", "Write query failed.");
		sqlite3_free(sql_error);
		ret = -1;
	}
	
	sqlite3_close(db);
	return ret;
}

static int acf_sqlite_read(struct ast_channel *chan, char *cmd, char *s, char *buf, size_t len)
{
	sqlite3 *db;
	struct acf_sqlite_query *query;
	char sql[2048] = "", varname[15];
	int x;
	char *file;
	char dbpath[PATH_MAX];
	char *sql_error;
	int rc, ret = 0;
	struct sqlite_function_cb *sfc = NULL;

	pbx_builtin_setvar_helper(chan, "SQL_ERROR", NULL);
	
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(field)[100];
	);
	
	AST_LIST_LOCK(&sqlite_queries);
	AST_LIST_TRAVERSE(&sqlite_queries, query, list) {
		if (query->acf) {
			if (!strcmp(query->acf->name, cmd)) {
				break;
			}
		} 
	}

	if (!query) {
		ast_log(LOG_ERROR, "No such function '%s'\n", cmd);
		pbx_builtin_setvar_helper(chan, "SQL_ERROR", "No matching function.");
		AST_LIST_UNLOCK(&sqlite_queries);
		return -1;
	}

	file = ast_strdup(query->file);

	if (file[0] == '/') {
		/* full path */
		strncpy(dbpath, file, sizeof(file));
	} else {
		sprintf(dbpath, "/etc/asterisk/%s", file);
	}

	if (sqlite3_open(dbpath, &db)) {
		ast_log(LOG_ERROR, "Failed to open database '%s' (check func_sqlite.conf)\n", dbpath);
		pbx_builtin_setvar_helper(chan, "SQL_ERROR", "Failed to open database.");
		AST_LIST_UNLOCK(&sqlite_queries);
		return -1;
	}
	
	sqlite3_busy_handler(db, sqlite_busy, NULL);

	AST_STANDARD_APP_ARGS(args, s);
	for (x = 0; x < args.argc; x++) {
		snprintf(varname, sizeof(varname), "ARG%d", x + 1);
		pbx_builtin_pushvar_helper(chan, varname, args.field[x]);
	}

	pbx_substitute_variables_helper(chan, query->sql_read, sql, sizeof(sql) - 1);

	/* Restore prior values */
	for (x = 0; x < args.argc; x++) {
		snprintf(varname, sizeof(varname), "ARG%d", x + 1);
		pbx_builtin_setvar_helper(chan, varname, NULL);
	}

	sfc = ast_malloc(sizeof(*sfc));
	sfc->record_count = 0;
	sfc->column_count = 0;
	sfc->escapecommas = ast_test_flag(query, OPT_ESCAPECOMMAS);
	sfc->buf = buf;
	sfc->len = len;

	AST_LIST_UNLOCK(&sqlite_queries);

	rc = sqlite3_exec(db, sql, sqlite_function_callback, sfc, &sql_error);
	if (rc != SQLITE_OK) {
		ast_log(LOG_ERROR, "Query failed to execute: %s\n", sql_error);
		pbx_builtin_setvar_helper(chan, "SQL_ERROR", "Read query failed to execute.");
		sqlite3_free(sql_error);
		ret = -1;
	}

	free(sfc);
	sqlite3_close(db);
	return ret;
}


static int init_acf_query(struct ast_config *cfg, char *catg, struct acf_sqlite_query **query)
{
	const char *tmp;

	if (!cfg || !catg) {
		return -1;
	}

	*query = ast_calloc(1, sizeof(struct acf_sqlite_query));
	if (! (*query))
		return -1;

	if ((tmp = ast_variable_retrieve(cfg, catg, "file"))) {
		ast_copy_string((*query)->file, tmp, sizeof((*query)->file));
	} else {
		return -1;
	}

	if ((tmp = ast_variable_retrieve(cfg, catg, "read"))) {
		ast_copy_string((*query)->sql_read, tmp, sizeof((*query)->sql_read));
	}

	if ((tmp = ast_variable_retrieve(cfg, catg, "write"))) {
		ast_copy_string((*query)->sql_write, tmp, sizeof((*query)->sql_write));
	}

	/* Allow escaping of embedded commas in fields to be turned off */
	ast_set_flag((*query), OPT_ESCAPECOMMAS);
	if ((tmp = ast_variable_retrieve(cfg, catg, "escapecommas"))) {
		if (ast_false(tmp))
			ast_clear_flag((*query), OPT_ESCAPECOMMAS);
	}

	(*query)->acf = ast_calloc(1, sizeof(struct ast_custom_function));
	if (! (*query)->acf) {
		free(*query);
		return -1;
	}

	if ((tmp = ast_variable_retrieve(cfg, catg, "prefix")) && !ast_strlen_zero(tmp)) {
		asprintf((char **)&((*query)->acf->name), "%s_%s", tmp, catg);
	} else {
		asprintf((char **)&((*query)->acf->name), "SQLITE_%s", catg);
	}

	if (!((*query)->acf->name)) {
		free((*query)->acf);
		free(*query);
		return -1;
	}

	asprintf((char **)&((*query)->acf->syntax), "%s(<arg1>[...[,<argN>]])", (*query)->acf->name);

	if (!((*query)->acf->syntax)) {
		free((char *)(*query)->acf->name);
		free((*query)->acf);
		free(*query);
		return -1;
	}

	(*query)->acf->synopsis = "Runs the referenced SQLite3 query with the specified arguments";
	if (!ast_strlen_zero((*query)->sql_read) && !ast_strlen_zero((*query)->sql_write)) {
		asprintf((char **)&((*query)->acf->desc),
					"Runs the following query, as defined in func_sqlite.conf, performing\n"
				   	"substitution of the arguments into the query as specified by ${ARG1},\n"
					"${ARG2}, ... ${ARGn}.  When setting the function, the values are provided\n"
					"either in whole as ${VALUE} or parsed as ${VAL1}, ${VAL2}, ... ${VALn}.\n"
					"\nRead:\n%s\n\nWrite:\n%s\n",
					(*query)->sql_read,
					(*query)->sql_write);
	} else if (!ast_strlen_zero((*query)->sql_read)) {
		asprintf((char **)&((*query)->acf->desc),
					"Runs the following query, as defined in func_sqlite.conf, performing\n"
				   	"substitution of the arguments into the query as specified by ${ARG1},\n"
					"${ARG2}, ... ${ARGn}.  This function may only be read, not set.\n\nSQL:\n%s\n",
					(*query)->sql_read);
	} else if (!ast_strlen_zero((*query)->sql_write)) {
		asprintf((char **)&((*query)->acf->desc),
					"Runs the following query, as defined in func_sqlite.conf, performing\n"
				   	"substitution of the arguments into the query as specified by ${ARG1},\n"
					"${ARG2}, ... ${ARGn}.  The values are provided either in whole as\n"
					"${VALUE} or parsed as ${VAL1}, ${VAL2}, ... ${VALn}.\n"
					"This function may only be set.\nSQL:\n%s\n",
					(*query)->sql_write);
	}

	/* Could be out of memory, or could be we have neither sql_read nor sql_write */
	if (! ((*query)->acf->desc)) {
		free((char *)(*query)->acf->syntax);
		free((char *)(*query)->acf->name);
		free((*query)->acf);
		free(*query);
		return -1;
	}

	if (ast_strlen_zero((*query)->sql_read)) {
		(*query)->acf->read = NULL;
	} else {
		(*query)->acf->read = acf_sqlite_read;
	}

	if (ast_strlen_zero((*query)->sql_write)) {
		(*query)->acf->write = NULL;
	} else {
		(*query)->acf->write = acf_sqlite_write;
	}

	return 0;
}

static int free_acf_query(struct acf_sqlite_query *query)
{
	if (query) {
		if (query->acf) {
			if (query->acf->name)
				free((char *)query->acf->name);
			if (query->acf->syntax)
				free((char *)query->acf->syntax);
			if (query->acf->desc)
				free((char *)query->acf->desc);
			free(query->acf);
		}
		free(query);
	}
	return 0;
}


/* SQLite 3 Manager Command */
static char *manager_query_synopsis = "Executes an SQLite3 query.";
	
static char manager_query_help[] =
"Description: The 'SQLite3' action may be used retrieve data from an\n"
"  SQLite3 database.\n"
"  \n"
"  Dbfile      - Required.  The file to open.  If the file is \n"
"                passed without a leading '/' the system presumes\n"
"                the file is located in the asterisk config directory.\n"
"  Query       - The query to execute\n";

static int manager_query(struct mansession *s, struct message *m)
{
	
	struct sqlite_manager_cb *mcb = NULL;
	char *dbfile = astman_get_header(m, "Dbfile");
	char *query = astman_get_header(m, "Query");
	char buffer[255];
	sqlite3 *db;
	int rc = 0;
	char *sql_error = NULL;
		
	if (ast_strlen_zero(dbfile)) {
		astman_send_error(s, m, "No database specified");
		return 0;
	}

	if (ast_strlen_zero(query)) {
		astman_send_error(s, m, "No query specified");
		return 0;
	}

	snprintf(buffer, sizeof(buffer), "/etc/asterisk/%s", dbfile);
	char *dbpath = buffer;
	
	
	if (sqlite3_open(dbpath, &db)) {
		astman_send_error(s, m, "Unable to open database file!\n");
		sqlite3_close(db);
		return 0;
	}

	sqlite3_busy_handler(db, sqlite_busy, NULL);
	
	mcb = ast_malloc(sizeof(*mcb));
	mcb->session = s;
	mcb->record_count = 0;
	
	rc = sqlite3_exec(db, query, sqlite_manager_callback, mcb, &sql_error);
	if (rc != SQLITE_OK) {
		astman_send_error(s, m, sql_error);
		sqlite3_free(sql_error);
	}
	astman_append(s, "\r\n\r\n");
	
	ast_free(mcb);
	sqlite3_close(db);
	return 0;
}

/* SQLite3 Dial Plan Function */


static int reload(void)
{
	int res = 0;
	struct ast_config *cfg;
	struct acf_sqlite_query *oldquery;
	char *catg;

	AST_LIST_LOCK(&sqlite_queries);

	while (!AST_LIST_EMPTY(&sqlite_queries)) {
		oldquery = AST_LIST_REMOVE_HEAD(&sqlite_queries, list);
		ast_custom_function_unregister(oldquery->acf);
		free_acf_query(oldquery);
	}

	cfg = ast_config_load(config);
	if (!cfg) {
		ast_log(LOG_WARNING, "Unable to load config for func_sqlite: %s\n", config);
		goto reload_out;
	}

	for (catg = ast_category_browse(cfg, NULL);
	     catg;
	     catg = ast_category_browse(cfg, catg)) {
		struct acf_sqlite_query *query = NULL;

		if (init_acf_query(cfg, catg, &query)) {
			ast_log(LOG_ERROR, "Cannot initialize query %s\n", catg);
		} else {
			AST_LIST_INSERT_HEAD(&sqlite_queries, query, list);
			ast_custom_function_register(query->acf);
		}
	}

	ast_config_destroy(cfg);

	reload_out:
		AST_LIST_UNLOCK(&sqlite_queries);
		return res;
}

static int unload_module(void)
{
	struct acf_sqlite_query *query;

	AST_LIST_LOCK(&sqlite_queries);
	while (!AST_LIST_EMPTY(&sqlite_queries)) {
		query = AST_LIST_REMOVE_HEAD(&sqlite_queries, list);
		ast_custom_function_unregister(query->acf);
		free_acf_query(query);
	}

	/* Allow any threads waiting for this lock to pass (avoids a race) */
	AST_LIST_UNLOCK(&sqlite_queries);
	AST_LIST_LOCK(&sqlite_queries);

	AST_LIST_UNLOCK(&sqlite_queries);
	return 0;
}

static int load_module(void)
{
	ast_manager_register2("Sqlite-Query", EVENT_FLAG_COMMAND, manager_query, manager_query_synopsis, manager_query_help);
	
	struct ast_config *cfg;
	char *catg;

	AST_LIST_LOCK(&sqlite_queries);

	cfg = ast_config_load(config);
	if (!cfg) {
		ast_log(LOG_NOTICE, "Unable to load config for func_sqlite: %s\n", config);
		AST_LIST_UNLOCK(&sqlite_queries);
		return AST_MODULE_LOAD_DECLINE;
	}

	for (catg = ast_category_browse(cfg, NULL);
	     catg;
	     catg = ast_category_browse(cfg, catg)) {
		struct acf_sqlite_query *query = NULL;

		if (init_acf_query(cfg, catg, &query)) {
			ast_log(LOG_ERROR, "Out of memory\n");
			free_acf_query(query);
		} else {
			AST_LIST_INSERT_HEAD(&sqlite_queries, query, list);
			ast_custom_function_register(query->acf);
		}
	}

	ast_config_destroy(cfg);

	AST_LIST_UNLOCK(&sqlite_queries);

	ast_log(LOG_NOTICE, "res_sqlite3 loaded.\n");
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS, "SQLite3 Resource",
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
	       );